/***********************************************************************
 *  File:  LANDEDGE.C
 *
 *  Purpose:  Calculates several different edge lengths at the landscape
 *	level:
 *  	   Landscape border:
 *		true edge; does not include edges between patches of 
 *		           type i bordering the landscape boundary, and
 *			   patches of type i in the landscape border
 *		all edge, regardless of whether it is true edge (total)
 *
 *	   No landscape border:
 *		edge length including landscape boundary (bound)
 *		edge length excluding landscape boundary (nobound)
 *
 *	   contrast weighted edge length
 * 
 *  Programmer:  Barbara Marks
 *
 *  Date:  January 1993
 *
 ************************************************************************/
#include <math.h>
#include "stats.h"


void landscape_edge_lengths ()
{

	int	edge1,edge2;
	short   value,neighbor,newval;
	short	xptr,yptr;
	short	i,j,k;
	char 	flag;



	read_image();

	total_edge = 0;
	total_true_edge = 0;
	total_bound_edge = 0;
	total_nobound_edge = 0;
	edcon = 0.0;

	for (i=0; i < num_wt; i++) {
	   for (j=0; j < num_wt; j++)
	      setint(edge,num_wt,i,j,0);
	}

/*
 *  Loop through all cells in the image
 */
	for (i=0; i < num_rows; i++) {
	   for (j=0; j < num_cols; j++) {

	      value = getshort (image,num_cols,i,j);

/*
 *  Only consider cells inside the landscape (values >= 0)
 */
	      if (value < 0) continue;

/*
 *  Check this cells four neighbors.  If the neighbor is a background
 *  cell (value of -999), or a landscape border edge, or its not equal 
 *  to the value of this cell, it is an edge.
 */
	      for (k=0; k < 4; k++) {
	         flag = FALSE;
		 xptr = j + xpos[k];
		 if (xptr < 0 || xptr >= num_cols) flag = TRUE;
		 yptr = i + ypos[k];
		 if (yptr < 0 || yptr >= num_rows) flag = TRUE;

/*
 *  This cell is on the image edge - increment the variable that
 *  includes the landscape boundary.
 */
		 if (flag) {
		    total_bound_edge ++;
/*
		    if (!land_border && use_bound) edcon++;
*/
		    if (use_bound) edcon ++;
		    continue;
		 }

/*
 *  Get the value of this neighbor.
 *
 *  For landscapes with borders, we don't want to count as an edge
 *  the landscape boundary if the adjacent class is the same as the
 *  boundary class.  "true" contains this modified edge length. 
 */ 
		 neighbor = getshort (image,num_cols,yptr,xptr);
		 if (neighbor == -1) continue;

		 if (neighbor != value) {
		    total_bound_edge++;
		    total_edge++;
/*
 *  Exclude from true_edge, any interior background edges, if the
 *  user requested not to treat the landscape boundary as an edge.
 */
		    if (neighbor == -999 && use_bound) 
		       total_true_edge ++;
		    if (neighbor != -999 && neighbor != -value)
		       total_true_edge ++;
		    if (neighbor != -999) total_nobound_edge++;

/*
 *  If requested, find the contrast weighted edge length
 */
		    if (contrast_indices) {
		       if (neighbor < 0 && neighbor != -999)
		           edcon += getfloat(contrast,num_wt,value-min_wt,
			         (-neighbor)-min_wt);
		       if (neighbor > 0)
			   edcon += getfloat(contrast,num_wt,value-min_wt,
				 neighbor-min_wt);
/*
		       if (!land_border && use_bound && neighbor == -999)
*/
		       if (use_bound && neighbor == -999)
		           edcon ++;
		    }

/*
 *  For contagion, increment the length between the patch type of
 *  of "value" and the patch type of "neighbor".
 */
		    if (neighbor == -999) continue;

		    newval = (short)fabs((double)neighbor);
		    edge1 = getint(edge,num_wt,value-min_wt,newval-min_wt);
		    edge2 = getint(edge,num_wt,newval-min_wt,value-min_wt);
		    if (land_border && neighbor != -value) {
		        if (value < newval)
		            edge1 ++;
  			else
		            edge2 ++;
		    }
		    if (!land_border) {
		        if (value < newval)
		            edge1 ++;
  		        else
		            edge2 ++;
		    }
		    setint(edge,num_wt,value-min_wt,newval-min_wt,edge1);
		    setint(edge,num_wt,newval-min_wt,value-min_wt,edge2);
		 }
	      }
/*
 *  Set the current cell to -1, so that it isn't evaluated again.
 */
              setshort (image,num_cols,i,j,-1);
	   }
	}
}
